#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
import unittest
from unittest import TestCase

from torchoutil.utils.collections import (
    flat_dict_of_dict,
    flat_list_of_list,
    intersect_lists,
    list_dict_to_dict_list,
    unflat_dict_of_dict,
    unflat_list_of_list,
)
from torchoutil.utils.type_checks import is_list_list_str, is_list_str


class TestFlatDict(TestCase):
    def test_example_1(self) -> None:
        x = {
            "a": 1,
            "b": {
                "a": 2,
                "b": 10,
            },
        }
        expected = {"a": 1, "b.a": 2, "b.b": 10}

        output = flat_dict_of_dict(x)
        self.assertDictEqual(output, expected)

    def test_example_2(self) -> None:
        x = {"a": ["hello", "world"], "b": 3}
        expected = {"a.0": "hello", "a.1": "world", "b": 3}

        output = flat_dict_of_dict(x, flat_iterables=True)
        self.assertDictEqual(output, expected)

    def test_example_3(self) -> None:
        x = {"a": {"b": 1}, "a.b": 2}
        expected = {"a.b": 2}

        output = flat_dict_of_dict(x)
        assert output == expected

    def test_example_4(self) -> None:
        x = {"a": {"b": 1}, "a.b": 2}
        self.assertRaises(ValueError, flat_dict_of_dict, x, overwrite=False)

    def test_unflat_dict_example_1(self) -> None:
        x = {
            "a.a": 1,
            "b.a": 2,
            "b.b": 3,
            "c": 4,
        }
        y = {"a": {"a": 1}, "b": {"a": 2, "b": 3}, "c": 4}

        y_hat = unflat_dict_of_dict(x)
        self.assertEqual(y_hat, y)


class TestListDictToDictList(TestCase):
    def test_example_1(self) -> None:
        lst = [{"a": 1, "b": 2}, {"a": 4, "b": 3, "c": 5}]
        output = list_dict_to_dict_list(lst, default_val=0)
        self.assertDictEqual(output, {"a": [1, 4], "b": [2, 3], "c": [0, 5]})

    def test_example_2(self) -> None:
        lst = [{"a": 1, "b": 2}, {"a": 4, "b": 3, "c": 5}]

        with self.assertRaises(ValueError):
            list_dict_to_dict_list(
                lst,
                default_val=None,
                key_mode="same",
            )

    def test_example_3(self) -> None:
        lst = [{"a": 1, "b": 2, "c": 3}, {"a": 11, "b": 22, "c": 33}]
        output = list_dict_to_dict_list(
            lst,
            default_val=None,
            key_mode="same",
        )
        self.assertDictEqual(output, {"a": [1, 11], "b": [2, 22], "c": [3, 33]})


class TestIntersectLists(TestCase):
    def test_example_1(self) -> None:
        input_ = [["a", "b", "b", "c"], ["c", "d", "b", "a"], ["b", "a", "a", "e"]]
        expected = ["a", "b"]

        output = intersect_lists(input_)
        self.assertListEqual(output, expected)


class TestFlatList(TestCase):
    def test_example_1(self) -> None:
        lst = [
            list(map(str, range(random.randint(0, 100))))
            for _ in range(random.randint(0, 10))
        ]
        for sublst in lst:
            random.shuffle(sublst)
        random.shuffle(lst)

        self.assertTrue(is_list_list_str(lst))

        flatten, sizes = flat_list_of_list(lst)

        self.assertTrue(is_list_str(flatten))
        self.assertEqual(len(lst), len(sizes))
        self.assertEqual(len(flatten), sum(sizes))

        unflat = unflat_list_of_list(flatten, sizes)

        self.assertTrue(is_list_list_str(unflat))
        self.assertEqual(len(lst), len(unflat))
        self.assertListEqual(lst, unflat)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import json
import logging
import math
import shutil
from pathlib import Path
from typing import Callable, List, Literal, Optional, TypeVar, Union

import torch
from torch import nn
from torch.utils.data.dataloader import DataLoader

from torchoutil.utils.data.dataloader import get_auto_num_cpus
from torchoutil.utils.data.dataset import SizedDatasetLike, TransformWrapper
from torchoutil.utils.pack.common import ATTRS_FNAME, CONTENT_DNAME, ContentMode
from torchoutil.utils.pack.dataset import PackedDataset
from torchoutil.utils.packaging import _TQDM_AVAILABLE
from torchoutil.utils.saving.common import to_builtin

if _TQDM_AVAILABLE:
    import tqdm


T = TypeVar("T")
U = TypeVar("U")


pylog = logging.getLogger(__name__)


@torch.inference_mode()
def pack_dataset(
    dataset: SizedDatasetLike[T],
    root: Union[str, Path],
    pre_transform: Optional[Callable[[T], U]] = None,
    batch_size: int = 32,
    num_workers: Union[int, Literal["auto"]] = "auto",
    overwrite: bool = False,
    content_mode: ContentMode = "item",
    custom_file_fmt: Union[None, str, Callable[[int], str]] = None,
    save_fn: Callable[[Union[U, List[U]], Path], None] = torch.save,
    subdir_size: Optional[int] = 100,
    verbose: int = 0,
) -> PackedDataset[U, U]:
    """Pack a dataset to pickle files.

    Here is an example how files are stored on disk for a dataset containing 1000 items:

    .. code-block:: text
        :caption:  Dataset folder tree example

        {root}
        ├── attributes.json
        └── data
            ├── 0
            |   └── 100 pt files
            ├── 1
            |   └── 100 pt files
            ├── ...
            |   └── ...
            └── 9
                └── 100 pt files

    Args:
        dataset: Dataset-like to pack.
        root: Directory to store pickled data.
        pre_transform: Transform to apply to each item before saving. defaults to None.
        batch_size: Batch size used by the dataloader. defaults to 32.
        num_workers: Number of workers used by the dataloader. defaults to "auto".
        overwrite: If True, overwrite all data in root directory. defaults to False.
        content_mode: Specify how the data should be stored.
            If "item", each dataset item will be stored in a separate file.
            If "batch", each dataset item will be stored in a batch file of size batch_size.
        custom_file_fmt: Custom file format.
            If None, defaults to "{{i:0{num_digits}d}}.pt".
            defaults to None.
        save_fn: Custom save function to save an item or a batch. defaults to torch.save.
        subdir_size: Optional number of files per folder.
            Using None will disable subdir an put all files in data/ folder.
            defaults to 100.
        verbose: Verbose level during packing. Higher value means more messages. defaults to 0.
    """

    # Check inputs
    if not isinstance(dataset, SizedDatasetLike):
        raise TypeError(
            f"Cannot pack to hdf a non-sized-dataset '{dataset.__class__.__name__}'."
        )
    if len(dataset) == 0:
        raise ValueError("Cannot pack to hdf an empty dataset.")

    root = Path(root).resolve()
    if root.exists() and not root.is_dir():
        raise RuntimeError(f"Item {root=} exists but it is not a file.")

    if num_workers == "auto":
        num_workers = get_auto_num_cpus()

    content_dpath = root.joinpath(CONTENT_DNAME)

    if content_dpath.is_dir():
        if not overwrite:
            raise ValueError(
                f"Cannot overwrite root data {str(content_dpath)}. Please remove it or use overwrite=True option."
            )
        shutil.rmtree(content_dpath)

    content_dpath.mkdir(parents=True, exist_ok=True)

    now = datetime.datetime.now()
    creation_date = now.strftime("%Y-%m-%d_%H-%M-%S")

    wrapped = TransformWrapper(dataset, pre_transform)

    loader = DataLoader(
        wrapped,  # type: ignore
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        collate_fn=nn.Identity(),
        drop_last=False,
        pin_memory=False,
    )

    if content_mode == "item":
        num_files = len(wrapped)
    elif content_mode == "batch":
        num_files = math.ceil(len(wrapped) / batch_size)
    else:
        raise ValueError(f"Invalid argument {content_mode=}.")

    if custom_file_fmt is None:
        num_digits = math.ceil(math.log10(num_files))
        file_fmt = f"{{:0{num_digits}d}}.pt".format
    elif isinstance(custom_file_fmt, str):
        file_fmt = custom_file_fmt.format
    else:
        file_fmt = custom_file_fmt

    fnames = [file_fmt(i) for i in range(num_files)]

    if subdir_size is not None and len(fnames) > subdir_size:
        num_subdirs = math.ceil(len(fnames) / subdir_size)
        num_digits = math.ceil(math.log10(num_subdirs))
        dir_fmt = f"{{:0{num_digits}d}}".format
        sub_dnames = [dir_fmt(i) for i in range(num_subdirs)]
        fnames = [
            str(Path(sub_dnames[i // subdir_size]).joinpath(fname))
            for i, fname in enumerate(fnames)
        ]

    fpaths = [content_dpath.joinpath(fname) for fname in fnames]

    if custom_file_fmt is not None or subdir_size is not None:
        unique_parents = dict.fromkeys(fpath.parent for fpath in fpaths)
        for parent in unique_parents:
            parent.mkdir(parents=True, exist_ok=True)

    if _TQDM_AVAILABLE and verbose >= 1:
        loader = tqdm.tqdm(loader)

    i = 0
    for batch_lst in loader:
        if content_mode == "item":
            for item in batch_lst:
                fpath = fpaths[i]
                save_fn(item, fpath)
                i += 1

        elif content_mode == "batch":
            fpath = fpaths[i]
            save_fn(batch_lst, fpath)
            i += 1

        else:
            raise ValueError(f"Invalid argument {content_mode=}.")

    if hasattr(dataset, "info"):
        info = dataset.info  # type: ignore
        info = to_builtin(info)
    else:
        info = {}

    if hasattr(dataset, "attrs"):
        source_attrs = dataset.attrs  # type: ignore
        source_attrs = to_builtin(source_attrs)
    else:
        source_attrs = {}

    attributes = {
        "source_dataset": wrapped.unwrap().__class__.__name__,
        "length": len(wrapped),
        "creation_date": creation_date,
        "batch_size": batch_size,
        "content_mode": content_mode,
        "content_dname": CONTENT_DNAME,
        "subdir_size": subdir_size,
        "info": info,
        "source_attrs": source_attrs,
        "num_files": len(fnames),
        "files": fnames,
    }

    attrs_fpath = root.joinpath(ATTRS_FNAME)
    with open(attrs_fpath, "w") as file:
        json.dump(attributes, file, indent="\t")

    pack = PackedDataset(root)
    return pack



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Collection of functions and modules to help development in PyTorch.
"""

__name__ = "torchoutil"
__author__ = "Étienne Labbé (Labbeti)"
__author_email__ = "labbeti.pub@gmail.com"
__license__ = "MIT"
__maintainer__ = "Étienne Labbé (Labbeti)"
__status__ = "Development"
__version__ = "0.4.0"


from torch import *

# Re-import for language servers
from . import hub as hub
from . import nn as nn
from . import optim as optim
from . import utils as utils
from .nn.functional import *
from .utils.type_checks import (
    is_numpy_scalar,
    is_python_scalar,
    is_scalar,
    is_torch_scalar,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Union

import torch
from torch import Tensor
from torch.types import Device

from torchoutil.nn.functional.get import get_device
from torchoutil.utils.packaging import _NUMPY_AVAILABLE

if not _NUMPY_AVAILABLE:
    ACCEPTED_NUMPY_DTYPES = ()

else:
    import numpy as np

    ACCEPTED_NUMPY_DTYPES = (
        np.float64,
        np.float32,
        np.float16,
        np.complex64,
        np.complex128,
        np.int64,
        np.int32,
        np.int16,
        np.int8,
        np.uint8,
        bool,
    )

    def to_numpy(
        x: Union[Tensor, np.ndarray, list],
        *,
        dtype: Union[str, np.dtype, None] = None,
    ) -> np.ndarray:
        """Convert input to numpy array."""
        if isinstance(x, Tensor):
            return tensor_to_numpy(x, dtype=dtype)
        else:
            return np.asarray(x, dtype=dtype)

    def tensor_to_numpy(
        x: Tensor,
        *,
        dtype: Union[str, np.dtype, None] = None,
    ) -> np.ndarray:
        """Convert PyTorch tensor to numpy array."""
        return x.cpu().numpy().astype(dtype=dtype)

    def numpy_to_tensor(
        x: np.ndarray,
        *,
        device: Device = None,
        dtype: Union[torch.dtype, None] = None,
    ) -> Tensor:
        """Convert numpy array to PyTorch tensor."""
        device = get_device(device)
        return torch.from_numpy(x).to(dtype=dtype, device=device)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, List, Literal, Optional, TypedDict

CONTENT_DNAME = "data"
ATTRS_FNAME = "attributes.json"

ContentMode = Literal["item", "batch"]


class PackedDatasetAttributes(TypedDict):
    source_dataset: str
    length: int
    creation_date: str
    batch_size: int
    content_mode: ContentMode
    content_dname: str
    subdir_size: Optional[int]
    info: Dict[str, Any]
    source_attrs: "PackedDatasetAttributes"
    num_files: int
    files: List[str]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Callable, Generic, Iterable, TypeVar, Union

import torch
from torch import Generator, Tensor, nn

from torchoutil.nn.functional.transform import (
    repeat_interleave_nd,
    resample_nearest_freqs,
    resample_nearest_rates,
    resample_nearest_steps,
    transform_drop,
)
from torchoutil.utils.collections import dump_dict

T = TypeVar("T")


class RepeatInterleaveNd(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.transform.repeat_interleave_nd`.
    """

    def __init__(self, repeats: int, dim: int) -> None:
        super().__init__()
        self.repeats = repeats
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return repeat_interleave_nd(x, self.repeats, self.dim)

    def extra_repr(self) -> str:
        return dump_dict(dict(repeats=self.repeats, dim=self.dim))


class ResampleNearestRates(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.transform.resample_nearest_rates`.
    """

    def __init__(
        self,
        rates: Union[float, Iterable[float]],
        dims: Union[int, Iterable[int]] = -1,
        round_fn: Callable[[Tensor], Tensor] = torch.floor,
    ) -> None:
        super().__init__()
        self.rates = rates
        self.dims = dims
        self.round_fn = round_fn

    def forward(self, x: Tensor) -> Tensor:
        return resample_nearest_rates(
            x,
            rates=self.rates,
            dims=self.dims,
            round_fn=self.round_fn,
        )

    def extra_repr(self) -> str:
        return dump_dict(dict(rates=self.rates, dims=self.dims))


class ResampleNearestFreqs(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.transform.resample_nearest_freqs`.
    """

    def __init__(
        self,
        orig_freq: int,
        new_freq: int,
        dims: Union[int, Iterable[int]] = -1,
        round_fn: Callable[[Tensor], Tensor] = torch.floor,
    ) -> None:
        super().__init__()
        self.orig_freq = orig_freq
        self.new_freq = new_freq
        self.dims = dims
        self.round_fn = round_fn

    def forward(self, x: Tensor) -> Tensor:
        return resample_nearest_freqs(
            x,
            orig_freq=self.orig_freq,
            new_freq=self.new_freq,
            dims=self.dims,
            round_fn=self.round_fn,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(orig_freq=self.orig_freq, new_freq=self.new_freq, dims=self.dims)
        )


class ResampleNearestSteps(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.transform.resample_nearest_steps`.
    """

    def __init__(
        self,
        steps: Union[float, Iterable[float]],
        dims: Union[int, Iterable[int]] = -1,
        round_fn: Callable[[Tensor], Tensor] = torch.floor,
    ) -> None:
        super().__init__()
        self.steps = steps
        self.dims = dims
        self.round_fn = round_fn

    def forward(self, x: Tensor) -> Tensor:
        return resample_nearest_steps(
            x,
            steps=self.steps,
            dims=self.dims,
            round_fn=self.round_fn,
        )

    def extra_repr(self) -> str:
        return dump_dict(dict(steps=self.steps, dims=self.dims))


class TransformDrop(Generic[T], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.transform.transform_drop`.
    """

    def __init__(
        self,
        transform: Callable[[T], T],
        p: float,
        generator: Union[int, Generator, None] = None,
    ) -> None:
        super().__init__()
        self.transform = transform
        self.p = p
        self.generator = generator

    def forward(self, x: T) -> T:
        return transform_drop(
            transform=self.transform,
            x=x,
            p=self.p,
            generator=self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(dict(p=self.p))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Literal, Optional, Union

import torch
from torch.types import Device

_DEVICE_CUDA_IF_AVAILABLE = "cuda_if_available"


def get_device(
    device: Union[Device, Literal["cuda_if_available"]] = _DEVICE_CUDA_IF_AVAILABLE,
) -> Optional[torch.device]:
    if device == _DEVICE_CUDA_IF_AVAILABLE:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    if isinstance(device, str):
        device = torch.device(device)
    return device



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

from torch import Tensor

from torchoutil.hub.registry import RegistryHub


class TestRegistryHub(TestCase):
    def test_cnext_register(self) -> None:
        register = RegistryHub(
            infos={
                "cnext_bl_70": {
                    "architecture": "ConvNeXt",
                    "url": "https://zenodo.org/record/8020843/files/convnext_tiny_465mAP_BL_AC_70kit.pth?download=1",
                    "hash_value": "0688ae503f5893be0b6b71cb92f8b428",
                    "hash_type": "md5",
                    "fname": "convnext_tiny_465mAP_BL_AC_70kit.pth",
                    "state_dict_key": "model",
                },
            },
        )

        model_name = "cnext_bl_70"
        register.download_file(model_name, force=False)
        state_dict = register.load_state_dict(model_name, offline=True, device="cpu")

        assert isinstance(state_dict, dict)
        assert all(isinstance(k, str) for k in state_dict.keys())
        assert all(isinstance(v, Tensor) for v in state_dict.values())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Helper functions for conversion between classes indices, onehot, names and probabilities for multiclass classification.
"""

from typing import Hashable, Iterable, List, Mapping, Optional, Sequence, TypeVar, Union

import torch
from torch import Tensor
from torch.nn import functional as F
from torch.types import Device

from torchoutil.nn.functional.get import get_device
from torchoutil.utils.type_checks import is_scalar

T = TypeVar("T")


def index_to_onehot(
    index: Union[Sequence[int], Tensor, Sequence],
    num_classes: int,
    *,
    padding_idx: Optional[int] = None,
    device: Device = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert indices of labels to onehot boolean encoding.

    Args:
        indices: List label indices.
            Can be a nested list of indices, but it should be convertible to Tensor.
        num_classes: Number maximal of unique classes.
        padding_idx: Class index to ignore. Output will contains only zeroes for this value. defaults to None.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    device = get_device(device)
    index = torch.as_tensor(index, device=device, dtype=torch.long)

    if padding_idx is not None:
        mask = index == padding_idx
        index = torch.where(mask, num_classes, index)
        num_classes += 1

    onehot: Tensor = F.one_hot(index, num_classes)
    onehot = onehot.to(dtype=dtype)

    if padding_idx is not None:
        onehot = onehot[..., :-1]

    return onehot


def index_to_name(
    index: Union[Sequence[int], Tensor, Sequence],
    idx_to_name: Union[Mapping[int, T], Sequence[T]],
) -> List[T]:
    """Convert indices of labels to names using a mapping.

    Args:
        indices: List of list of label indices.
        idx_to_name: Mapping to convert a class index to its name.
    """

    def index_to_name_impl(x) -> Union[T, list]:
        if is_scalar(x):
            return idx_to_name[x]  # type: ignore
        elif isinstance(x, Iterable):
            return [index_to_name_impl(xi) for xi in x]
        else:
            raise ValueError(
                f"Invalid argument {x=}. (not present in idx_to_name and not an iterable type)"
            )

    name = index_to_name_impl(index)
    return name  # type: ignore


def onehot_to_index(
    onehot: Tensor,
    *,
    dim: int = -1,
) -> List[int]:
    """Convert onehot boolean encoding to indices of labels.

    Args:
        onehot: Onehot labels encoded as 2D matrix.
    """
    onehot = onehot.int()
    index = onehot.argmax(dim=dim)
    index = index.tolist()
    return index


def onehot_to_name(
    onehot: Tensor,
    idx_to_name: Union[Mapping[int, T], Sequence[T]],
    *,
    dim: int = -1,
) -> List[T]:
    """Convert onehot boolean encoding to names using a mapping.

    Args:
        onehot: Onehot labels encoded as 2D matrix.
        idx_to_name: Mapping to convert a class index to its name.
    """
    indices = onehot_to_index(onehot, dim=dim)
    names = index_to_name(indices, idx_to_name)
    return names


def name_to_index(
    name: List[T],
    idx_to_name: Union[Mapping[int, T], Sequence[T]],
) -> List[int]:
    """Convert names to indices of labels.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
    """
    if isinstance(idx_to_name, Mapping):
        name_to_idx = {name: idx for idx, name in idx_to_name.items()}
    else:
        name_to_idx = {name: idx for idx, name in enumerate(idx_to_name)}
    del idx_to_name

    def name_to_index_impl(x) -> Union[T, list]:
        if isinstance(x, Hashable) and x in name_to_idx:
            return name_to_idx[x]  # type: ignore
        elif isinstance(x, Iterable):
            return [name_to_index_impl(xi) for xi in x]
        else:
            raise ValueError(
                f"Invalid argument {x=}. (not present in name_to_idx and not an iterable type)"
            )

    index = name_to_index_impl(name)
    return index  # type: ignore


def name_to_onehot(
    name: List[T],
    idx_to_name: Union[Mapping[int, T], Sequence[T]],
    *,
    device: Device = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert names to onehot boolean encoding.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    index = name_to_index(name, idx_to_name)
    onehot = index_to_onehot(index, len(idx_to_name), device=device, dtype=dtype)
    return onehot


def probs_to_index(
    probs: Tensor,
    *,
    dim: int = -1,
) -> List[int]:
    """Convert matrix of probabilities to indices of labels.

    Args:
        probs: Output probabilities for each classes.
    """
    index = probs.argmax(dim=dim)
    index = index.tolist()
    return index


def probs_to_onehot(
    probs: Tensor,
    *,
    dim: int = -1,
    device: Device = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert matrix of probabilities to onehot boolean encoding.

    Args:
        probs: Output probabilities for each classes.
    """
    if device is None:
        device = probs.device
    indices = probs_to_index(probs, dim=dim)
    onehot = index_to_onehot(indices, probs.shape[-1], device=device, dtype=dtype)
    return onehot


def probs_to_name(
    probs: Tensor,
    idx_to_name: Union[Mapping[int, T], Sequence[T]],
    *,
    dim: int = -1,
) -> List[T]:
    """Convert matrix of probabilities to labels names.

    Args:
        probs: Output probabilities for each classes.
        idx_to_name: Mapping to convert a class index to its name.
    """
    indices = probs_to_index(probs, dim=dim)
    names = index_to_name(indices, idx_to_name)
    return names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Helper functions for conversion between classes indices, multihot, names and probabilities for multilabel classification.
"""

from typing import Hashable, Iterable, List, Mapping, Optional, Sequence, TypeVar, Union

import torch
from torch import Tensor
from torch.types import Device

from torchoutil.nn.functional.get import get_device
from torchoutil.nn.functional.pad import pad_and_stack_rec
from torchoutil.utils.type_checks import is_scalar, is_sequence_bool, is_sequence_int

T = TypeVar("T")


def indices_to_multihot(
    indices: Union[Sequence[Union[Sequence[int], Tensor]], Tensor],
    num_classes: int,
    *,
    padding_idx: Optional[int] = None,
    device: Device = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert indices of labels to multihot boolean encoding.

    Args:
        indices: List of list of label indices. Values should be integers in range [0..num_classes-1]
        num_classes: Number maximal of unique classes.
        padding_idx: Optional pad value to ignore.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    if device is None and isinstance(indices, Tensor):
        device = indices.device
    else:
        device = get_device(device)

    def _indices_to_multihot_impl(x) -> Tensor:
        if isinstance(x, Tensor) and not _is_valid_indices(x):
            raise ValueError(f"Invalid argument shape {x.shape=}.")

        if (isinstance(x, Tensor) and x.ndim == 1) or is_sequence_int(x):
            x = torch.as_tensor(x, dtype=torch.long, device=device)

            if padding_idx is not None:
                x = torch.where(x == padding_idx, num_classes, x)
                target_num_classes = num_classes + 1
            else:
                target_num_classes = num_classes

            multihot = torch.full(
                (target_num_classes,), False, dtype=dtype, device=device
            )
            multihot.scatter_(0, x, True)

            if padding_idx is not None:
                multihot = multihot[..., :-1]

            return multihot

        elif isinstance(x, Iterable):
            result = [_indices_to_multihot_impl(xi) for xi in x]
            return torch.stack(result)

        else:
            raise ValueError(f"Invalid argument {x=}.")

    multihot = _indices_to_multihot_impl(indices)
    return multihot


def indices_to_names(
    indices: Union[Sequence[Union[Sequence[int], Tensor]], Tensor],
    idx_to_name: Mapping[int, T],
    *,
    padding_idx: Optional[int] = None,
) -> List[List[T]]:
    """Convert indices of labels to names using a mapping.

    Args:
        indices: List of list of label indices.
        idx_to_name: Mapping to convert a class index to its name.
        padding_idx: Optional pad value to ignore.
    """

    def _indices_to_names_impl(x) -> Union[int, list]:
        if is_scalar(x):
            return idx_to_name[x]  # type: ignore
        elif isinstance(x, Iterable):
            return [
                _indices_to_names_impl(xi)
                for xi in x
                if padding_idx is None or not is_scalar(xi) or xi != padding_idx
            ]
        else:
            raise ValueError(
                f"Invalid argument {x=}. (not present in idx_to_name and not an iterable type)"
            )

    if not isinstance(indices, Iterable):
        raise TypeError(f"Invalid argument {indices=}. (not an iterable)")

    names = _indices_to_names_impl(indices)
    return names  # type: ignore


def multihot_to_indices(
    multihot: Union[Tensor, Sequence[Tensor], Sequence[Sequence[bool]]],
    *,
    padding_idx: Optional[int] = None,
) -> List[List[int]]:
    """Convert multihot boolean encoding to indices of labels.

    Args:
        multihot: Multihot labels encoded as 2D matrix.
    """
    if is_scalar(multihot) or (
        isinstance(multihot, Tensor) and not _is_valid_indices(multihot)
    ):
        raise ValueError(
            f"Invalid argument shape {multihot=}. (expected at least 1 dimension and the first axis should be > 0)"
        )

    def _multihot_to_indices_impl(
        x: Union[Tensor, Sequence],
    ) -> list:
        if (isinstance(x, Tensor) and x.ndim == 1) or is_sequence_bool(x):
            x = torch.as_tensor(x, dtype=torch.bool)
            preds = torch.where(x)[0].tolist()
            return preds
        elif (isinstance(x, Tensor) and x.ndim > 1) or isinstance(x, Sequence):
            preds = [_multihot_to_indices_impl(multihot_i) for multihot_i in x]
            if padding_idx is not None:
                preds = pad_and_stack_rec(preds, padding_idx)
                preds = preds.tolist()
            return preds

        else:
            raise ValueError(
                f"Invalid argument {x=}. (not present in idx_to_name and not an iterable type)"
            )

    return _multihot_to_indices_impl(multihot)


def multihot_to_names(
    multihot: Union[Tensor, Sequence[Tensor], Sequence[Sequence[bool]]],
    idx_to_name: Mapping[int, T],
) -> List[List[T]]:
    """Convert multihot boolean encoding to names using a mapping.

    Args:
        multihot: Multihot labels encoded as 2D matrix.
        idx_to_name: Mapping to convert a class index to its name.
    """
    indices = multihot_to_indices(multihot)
    names = indices_to_names(indices, idx_to_name)
    return names


def names_to_indices(
    names: List[List[T]],
    idx_to_name: Mapping[int, T],
) -> List[List[int]]:
    """Convert names to indices of labels.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
    """
    name_to_idx = {name: idx for idx, name in idx_to_name.items()}
    del idx_to_name

    def _names_to_indices_impl(x) -> Union[int, list]:
        if isinstance(x, Hashable) and x in name_to_idx:
            return name_to_idx[x]  # type: ignore
        elif isinstance(x, Iterable):
            return [_names_to_indices_impl(xi) for xi in x]
        else:
            raise ValueError(
                f"Invalid argument {x=}. (not present in idx_to_name and not an iterable type)"
            )

    indices = _names_to_indices_impl(names)
    return indices  # type: ignore


def names_to_multihot(
    names: List[List[T]],
    idx_to_name: Mapping[int, T],
    *,
    device: Device = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert names to multihot boolean encoding.

    Args:
        names: List of list of label names.
        idx_to_name: Mapping to convert a class index to its name.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    indices = names_to_indices(names, idx_to_name)
    multihot = indices_to_multihot(
        indices,
        len(idx_to_name),
        device=device,
        dtype=dtype,
    )
    return multihot


def probs_to_indices(
    probs: Tensor,
    threshold: Union[float, Sequence[float], Tensor],
    *,
    padding_idx: Optional[int] = None,
) -> List[List[int]]:
    """Convert matrix of probabilities to indices of labels.

    Args:
        probs: Output probabilities for each classes.
        threshold: Threshold(s) to binarize probabilities. Can be a scalar or a sequence of (num_classes,) thresholds.
    """
    multihot = probs_to_multihot(probs, threshold)
    indices = multihot_to_indices(multihot, padding_idx=padding_idx)
    return indices


def probs_to_multihot(
    probs: Tensor,
    threshold: Union[float, Sequence[float], Tensor],
    *,
    device: Device = None,
    dtype: Union[torch.dtype, None] = torch.bool,
) -> Tensor:
    """Convert matrix of probabilities to multihot boolean encoding.

    Args:
        probs: Output probabilities for each classes.
        threshold: Threshold(s) to binarize probabilities. Can be a scalar or a sequence of (num_classes,) thresholds.
        device: PyTorch device of the output tensor.
        dtype: PyTorch DType of the output tensor.
    """
    if probs.ndim != 2:
        raise ValueError(
            "Invalid argument probs. (expected a batch of probabilities of shape (N, n_classes))."
        )

    num_classes = probs.shape[-1]
    if device is None:
        device = probs.device
    else:
        device = get_device(device)

    if not isinstance(threshold, Tensor):
        threshold = torch.as_tensor(threshold, dtype=torch.float, device=device)
    else:
        threshold = threshold.to(device=device)

    if threshold.ndim == 0:
        threshold = threshold.repeat(num_classes)
    elif threshold.ndim == 1:
        if threshold.shape[0] != num_classes:
            raise ValueError(
                f"Invalid argument threshold. (number of thresholds is {threshold.shape[0]} but found {num_classes} classes)"
            )
    else:
        raise ValueError(
            f"Invalid number of dimensions in input threshold. (found {threshold.shape=} but expected 0-d or 1-d tensor)"
        )

    multihot = probs >= threshold
    multihot = multihot.to(dtype=dtype)
    return multihot


def probs_to_names(
    probs: Tensor,
    threshold: Union[float, Sequence[float], Tensor],
    idx_to_name: Mapping[int, T],
) -> List[List[T]]:
    """Convert matrix of probabilities to labels names.

    Args:
        probs: Output probabilities for each classes.
        threshold: Threshold(s) to binarize probabilities. Can be a scalar or a sequence of (num_classes,) thresholds.
        idx_to_name: Mapping to convert a class index to its name.
    """
    indices = probs_to_indices(probs, threshold)
    names = indices_to_names(indices, idx_to_name)
    return names


def _is_valid_indices(x: Tensor, dim: int = -1) -> bool:
    """Returns True if tensor has valid shape for indices_to_multihot."""
    shape = list(x.shape)
    dim = dim % len(shape)
    try:
        idx = shape.index(0)
        return idx == dim
    except ValueError:
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from torchoutil.utils.type_checks import is_mapping_str

K = TypeVar("K")
T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
W = TypeVar("W")

KEY_MODES = ("same", "intersect", "union")
KeyMode = Literal["intersect", "same", "union"]


def sorted_dict(
    x: Mapping[K, V],
    /,
    *,
    key: Optional[Callable[[K], Any]] = None,
    reverse: bool = False,
) -> Dict[K, V]:
    return {k: x[k] for k in sorted(x.keys(), key=key, reverse=reverse)}


@overload
def list_dict_to_dict_list(
    lst: Sequence[Mapping[K, V]],
    key_mode: Literal["intersect", "same"],
    default_val: Any = None,
) -> Dict[K, List[V]]:
    ...


@overload
def list_dict_to_dict_list(
    lst: Sequence[Mapping[K, V]],
    key_mode: Literal["union"] = "union",
    default_val: W = None,
) -> Dict[K, List[Union[V, W]]]:
    ...


def list_dict_to_dict_list(
    lst: Sequence[Mapping[K, V]],
    key_mode: KeyMode = "union",
    default_val: W = None,
) -> Dict[K, List[Union[V, W]]]:
    """Convert list of dicts to dict of lists.

    Args:
        lst: The list of dict to merge.
        key_mode: Can be "same" or "intersect".
            If "same", all the dictionaries must contains the same keys otherwise a ValueError will be raised.
            If "intersect", only the intersection of all keys will be used in output.
            If "union", the output dict will contains the union of all keys, and the missing value will use the argument default_val.
        default_val: Default value of an element when key_mode is "union". defaults to None.
    """
    if len(lst) <= 0:
        return {}

    keys = set(lst[0].keys())
    if key_mode == "same":
        invalids = [list(item.keys()) for item in lst[1:] if keys != set(item.keys())]
        if len(invalids):
            raise ValueError(
                f"Invalid keys with {key_mode=}. (with {keys=} and {invalids=})"
            )
    elif key_mode == "intersect":
        keys = intersect_lists([item.keys() for item in lst])
    elif key_mode == "union":
        keys = union_lists([item.keys() for item in lst])
    else:
        raise ValueError(
            f"Invalid argument key_mode={key_mode}. (expected one of {KEY_MODES})"
        )

    return {key: [item.get(key, default_val) for item in lst] for key in keys}


@overload
def dict_list_to_list_dict(
    dic: Mapping[T, Sequence[U]],
    key_mode: Literal["same", "intersect"],
    default_val: Any = None,
) -> List[Dict[T, U]]:
    ...


@overload
def dict_list_to_list_dict(
    dic: Mapping[T, Sequence[U]],
    key_mode: Literal["union"] = "union",
    default_val: W = None,
) -> List[Dict[T, Union[U, W]]]:
    ...


def dict_list_to_list_dict(
    dic: Mapping[T, Sequence[U]],
    key_mode: KeyMode = "union",
    default_val: W = None,
) -> List[Dict[T, Union[U, W]]]:
    """Convert dict of lists with same sizes to list of dicts.

    Example 1
    ----------
    ```
    >>> dic = {"a": [1, 2], "b": [3, 4]}
    >>> dict_list_to_list_dict(dic)
    ... [{"a": 1, "b": 3}, {"a": 2, "b": 4}]
    ```
    """
    if len(dic) == 0:
        return []

    lengths = [len(seq) for seq in dic.values()]
    if key_mode == "same":
        if not all_eq(lengths):
            raise ValueError("Invalid sequences for batch.")
        length = lengths[0]
    elif key_mode == "intersect":
        length = min(lengths)
    elif key_mode == "union":
        length = max(lengths)
    else:
        raise ValueError(
            f"Invalid argument key_mode={key_mode}. (expected one of {KEY_MODES})"
        )

    result = [
        {k: (v[i] if i < len(v) else default_val) for k, v in dic.items()}
        for i in range(length)
    ]
    return result


def intersect_lists(lst_of_lst: Sequence[Iterable[T]]) -> List[T]:
    """Performs intersection of elements in lists (like set intersection), but keep their original order."""
    if len(lst_of_lst) <= 0:
        return []
    out = list(dict.fromkeys(lst_of_lst[0]))
    for lst_i in lst_of_lst[1:]:
        out = [name for name in out if name in lst_i]
        if len(out) == 0:
            break
    return out


def union_lists(lst_of_lst: Iterable[Iterable[T]]) -> List[T]:
    """Performs union of elements in lists (like set union), but keep their original order."""
    out = {}
    for lst_i in lst_of_lst:
        out.update(dict.fromkeys(lst_i))
    out = list(out)
    return out


def dump_dict(
    dic: Mapping[str, T],
    /,
    join: str = ", ",
    fmt: str = "{key}={value}",
    ignore_lst: Iterable[T] = (),
) -> str:
    """Dump dictionary to string."""
    ignore_lst = dict.fromkeys(ignore_lst)
    result = join.join(
        fmt.format(key=key, value=value)
        for key, value in dic.items()
        if value not in ignore_lst
    )
    return result


def pass_filter(
    x: T,
    include: Optional[Iterable[T]] = None,
    exclude: Optional[Iterable[T]] = None,
) -> bool:
    """Returns True if name in include set and not in exclude set."""
    if include is not None and exclude is not None:
        return (x in include) and (x not in exclude)
    if include is not None:
        return x in include
    elif exclude is not None:
        return x not in exclude
    else:
        return True


def filter_iterable(
    it: Iterable[T],
    include: Optional[Iterable[T]] = None,
    exclude: Optional[Iterable[T]] = None,
) -> List[T]:
    return [item for item in it if pass_filter(item, include, exclude)]


def all_eq(it: Iterable[T], eq_fn: Optional[Callable[[T, T], bool]] = None) -> bool:
    """Returns true if all elements in inputs are equal."""
    it = list(it)
    first = it[0]
    if eq_fn is None:
        return all(first == elt for elt in it)
    else:
        return all(eq_fn(first, elt) for elt in it)


def all_ne(it: Iterable[T], ne_fn: Optional[Callable[[T, T], bool]] = None) -> bool:
    """Returns true if all elements in inputs are differents."""
    it = list(it)
    if ne_fn is None:
        return all(
            it[i] != it[j] for i in range(len(it)) for j in range(i + 1, len(it))
        )
    else:
        return all(
            ne_fn(it[i], it[j]) for i in range(len(it)) for j in range(i + 1, len(it))
        )


def flat_dict_of_dict(
    nested_dic: Mapping[str, Any],
    *,
    sep: str = ".",
    flat_iterables: bool = False,
    overwrite: bool = True,
) -> Dict[str, Any]:
    """Flat a nested dictionary.

    Example 1
    ---------
    ```
    >>> dic = {
    ...     "a": 1,
    ...     "b": {
    ...         "a": 2,
    ...         "b": 10,
    ...     },
    ... }
    >>> flat_dict_of_dict(dic)
    ... {"a": 1, "b.a": 2, "b.b": 10}
    ```

    Example 2
    ---------
    ```
    >>> dic = {"a": ["hello", "world"], "b": 3}
    >>> flat_dict_of_dict(dic, flat_iterables=True)
    ... {"a.0": "hello", "a.1": "world", "b": 3}
    ```

    Args:
        nested_dict: Nested mapping containing sub-mappings or iterables.
        sep: Separators between keys.
        flat_iterables: If True, flat iterable and use index as key.
        overwrite: If True, overwrite duplicated keys in output. Otherwise duplicated keys will raises a ValueError.
    """

    def _flat_dict_of_dict_impl(nested_dic: Mapping[str, Any]) -> Dict[str, Any]:
        output = {}
        for k, v in nested_dic.items():
            if is_mapping_str(v):
                v = _flat_dict_of_dict_impl(v)
                v = {f"{k}{sep}{kv}": vv for kv, vv in v.items()}
                output.update(v)

            elif flat_iterables and isinstance(v, Iterable) and not isinstance(v, str):
                v = {f"{i}": vi for i, vi in enumerate(v)}
                v = _flat_dict_of_dict_impl(v)
                v = {f"{k}{sep}{kv}": vv for kv, vv in v.items()}
                output.update(v)

            elif overwrite or k not in output:
                output[k] = v

            else:
                raise ValueError(
                    f"Ambiguous flatten dict with key '{k}'. (with value '{v}')"
                )
        return output

    return _flat_dict_of_dict_impl(nested_dic)


def unflat_dict_of_dict(dic: Mapping[str, Any], sep: str = ".") -> Dict[str, Any]:
    """Unflat a dictionary.

    Example 1
    ----------
    ```
    >>> dic = {
        "a.a": 1,
        "b.a": 2,
        "b.b": 3,
        "c": 4,
    }
    >>> unflat_dict_of_dict(dic)
    ... {"a": {"a": 1}, "b": {"a": 2, "b": 3}, "c": 4}
    ```
    """
    output = {}
    for k, v in dic.items():
        if sep not in k:
            output[k] = v
        else:
            idx = k.index(sep)
            k, kk = k[:idx], k[idx + 1 :]
            if k not in output:
                output[k] = {}
            elif not isinstance(output[k], Mapping):
                raise ValueError(
                    f"Invalid dict argument. (found keys {k} and {k}{sep}{kk})"
                )

            output[k][kk] = v

    output = {
        k: (unflat_dict_of_dict(v) if isinstance(v, Mapping) else v)
        for k, v in output.items()
    }
    return output


def flat_list_of_list(lst: Iterable[Sequence[T]]) -> Tuple[List[T], List[int]]:
    """Return a flat version of the input list of sublists with each sublist size."""
    flatten_lst = [elt for sublst in lst for elt in sublst]
    sizes = [len(sents) for sents in lst]
    return flatten_lst, sizes


def unflat_list_of_list(
    flatten_lst: Sequence[T],
    sizes: Iterable[int],
) -> List[List[T]]:
    """Unflat a list to a list of sublists of given sizes."""
    lst = []
    start = 0
    stop = 0
    for count in sizes:
        stop += count
        lst.append(flatten_lst[start:stop])
        start = stop
    return lst


@overload
def unzip(lst: Iterable[Tuple[T]]) -> Tuple[List[T]]:
    ...


@overload
def unzip(lst: Iterable[Tuple[T, U]]) -> Tuple[List[T], List[U]]:
    ...


@overload
def unzip(lst: Iterable[Tuple[T, U, V]]) -> Tuple[List[T], List[U], List[V]]:
    ...


@overload
def unzip(
    lst: Iterable[Tuple[T, U, V, W]]
) -> Tuple[List[T], List[U], List[V], List[W]]:
    ...


def unzip(lst):
    """Invert zip() function.

    .. code-block:: python
        :caption:  Example

        >>> lst1 = [1, 2, 3, 4]
        >>> lst2 = [5, 6, 7, 8]
        >>> lst_zipped = list(zip(lst1, lst2))
        >>> lst_zipped
        ... [(1, 5), (2, 6), (3, 7), (4, 8)]
        >>> unzip(lst_zipped)
        ... ([1, 2, 3, 4], [5, 6, 7, 8])
    """
    return tuple(map(list, zip(*lst)))


def prod(x: Iterable[T], /, start: T = 1) -> T:
    result = copy.copy(start)
    for xi in x:
        result = result * xi
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import hashlib
from pathlib import Path
from typing import Literal, Optional, Union

from torch.hub import download_url_to_file

DEFAULT_CHUNK_SIZE = 256 * 1024**2  # 256 MiB
HASH_TYPES = ("sha256", "md5")
HashType = Literal["sha256", "md5"]


def hash_file(
    fpath: Union[str, Path],
    hash_type: HashType,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
) -> str:
    """Return the hash value for a file.

    BASED ON https://github.com/pytorch/audio/blob/v0.13.0/torchaudio/datasets/utils.py#L110
    """
    if hash_type == "sha256":
        hasher = hashlib.sha256()
    elif hash_type == "md5":
        hasher = hashlib.md5()
    else:
        raise ValueError(
            f"Invalid argument hash_type={hash_type}. (expected one of {HASH_TYPES})"
        )

    with open(fpath, "rb") as file:
        while True:
            chunk = file.read(chunk_size)
            if not chunk:
                break
            hasher.update(chunk)

    hash_value = hasher.hexdigest()
    return hash_value


def download_file(
    url: str,
    fpath: Union[str, Path],
    hash_prefix: Optional[str] = None,
    make_intermediate: bool = False,
    verbose: int = 0,
) -> Path:
    fpath = Path(fpath)
    if make_intermediate:
        dpath = fpath.parent
        dpath.mkdir(parents=True, exist_ok=True)

    download_url_to_file(url, fpath, hash_prefix=hash_prefix, progress=verbose > 0)
    return fpath



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from .collate import AdvancedCollateDict, CollateDict
from .dataloader import get_auto_num_cpus
from .dataset import EmptyDataset, SizedDatasetLike, TransformWrapper
from .split import balanced_monolabel_split, random_split



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Union

import torch
from torch import Tensor, nn
from torch.types import Device

from torchoutil.nn.functional.numpy import numpy_to_tensor, tensor_to_numpy, to_numpy
from torchoutil.utils.packaging import _NUMPY_AVAILABLE

if not _NUMPY_AVAILABLE:

    class np:
        dtype = Any
        ndarray = Any

else:
    import numpy as np


class ToNumpy(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.numpy.to_numpy`.
    """

    def __init__(self, *, dtype: Union[str, np.dtype, None] = None) -> None:
        super().__init__()
        self.dtype = dtype

    def forward(self, x: Union[Tensor, np.ndarray, list]) -> np.ndarray:
        return to_numpy(x, dtype=self.dtype)


class TensorToNumpy(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.numpy.tensor_to_numpy`.
    """

    def __init__(self, *, dtype: Union[str, np.dtype, None] = None) -> None:
        super().__init__()
        self.dtype = dtype

    def forward(self, x: Union[Tensor, np.ndarray, list]) -> np.ndarray:
        return tensor_to_numpy(x, dtype=self.dtype)


class NumpyToTensor(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.numpy.numpy_to_tensor`.
    """

    def __init__(
        self,
        *,
        device: Device = None,
        dtype: Union[torch.dtype, None] = None,
    ) -> None:
        super().__init__()
        self.device = device
        self.dtype = dtype

    def forward(self, x: np.ndarray) -> Tensor:
        return numpy_to_tensor(x, dtype=self.dtype, device=self.device)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
from typing import Callable, Iterable, List, TypeVar, Union

import torch
from torch import Generator, Tensor

from torchoutil.utils.type_checks import is_scalar

T = TypeVar("T")
U = TypeVar("U")


def repeat_interleave_nd(x: Tensor, repeats: int, dim: int = 0) -> Tensor:
    """Generalized version of torch.repeat_interleave for N >= 1 dimensions.
    The output size will be (..., D*repeats, ...), where D is the size of the dimension of the dim argument.

    Args:
        x: Any tensor of shape (..., D, ...) with at least 1 dim.
        repeats: Number of repeats.
        dim: The dimension to repeat. defaults to 0.

    Examples::
    ----------
        >>> x = torch.as_tensor([[0, 1, 2, 3], [4, 5, 6, 7]])
        >>> repeat_interleave_nd(x, n=2, dim=0)
        tensor([[0, 1, 2, 3],
                [0, 1, 2, 3],
                [4, 5, 6, 7],
                [4, 5, 6, 7]])
    """
    if x.ndim == 0:
        raise ValueError(
            f"Function repeat_interleave_nd does not supports 0-d tensors. (found {x.ndim=} == 0)"
        )

    dim = dim % x.ndim
    x = x.unsqueeze(dim=dim + 1)
    shape = list(x.shape)
    shape[dim + 1] = repeats
    x = x.expand(*shape)
    x = x.flatten(dim, dim + 1)
    return x


def resample_nearest_rates(
    x: Tensor,
    rates: Union[float, Iterable[float]],
    *,
    dims: Union[int, Iterable[int]] = -1,
    round_fn: Callable[[Tensor], Tensor] = torch.floor,
) -> Tensor:
    """Nearest neigbour resampling using tensor slices.

    Args:
        x: Input tensor.
        rate: The reduction factor of each axis, e.g. a factor of 0.5 will divide the input axes by 2.
        dims: Dimensions to apply resampling. defaults to -1.
        round_fn: Rounding function to compute sub-indices. defaults to torch.floor.
    """
    if isinstance(dims, int):
        dims = [dims]
    else:
        dims = list(dims)

    if is_scalar(rates):
        steps = [1.0 / rates] * len(dims)
    else:
        steps = [1.0 / rate for rate in rates]

    return resample_nearest_steps(
        x,
        steps,
        dims=dims,
        round_fn=round_fn,
    )


def resample_nearest_freqs(
    x: Tensor,
    orig_freq: int,
    new_freq: int,
    *,
    dims: Union[int, Iterable[int]] = -1,
    round_fn: Callable[[Tensor], Tensor] = torch.floor,
) -> Tensor:
    return resample_nearest_steps(
        x,
        orig_freq / new_freq,
        dims=dims,
        round_fn=round_fn,
    )


def resample_nearest_steps(
    x: Tensor,
    steps: Union[float, Iterable[float]],
    *,
    dims: Union[int, Iterable[int]] = -1,
    round_fn: Callable[[Tensor], Tensor] = torch.floor,
) -> Tensor:
    if isinstance(dims, int):
        dims = [dims]
    else:
        dims = list(dims)

    if is_scalar(steps):
        steps = [steps] * len(dims)
    else:
        steps = list(steps)  # type: ignore
        if len(steps) != len(dims):
            raise ValueError(f"Invalid arguments sizes {len(steps)=} != {len(dims)}.")

    slices: List[Union[slice, Tensor]] = [slice(None)] * x.ndim

    for dim, step in zip(dims, steps):
        length = x.shape[dim]
        indexes = torch.arange(0, length, step)
        indexes = round_fn(indexes).long().clamp(min=0, max=length - 1)
        slices[dim] = indexes

    x = x[slices]
    return x


def transform_drop(
    transform: Callable[[T], T],
    x: T,
    p: float,
    generator: Union[int, Generator, None] = None,
) -> T:
    """Apply a transform on a tensor with a probability of p.

    Args:
        transform: Transform to apply.
        x: Argument of the transform.
        p: Probability p to apply the transform. Cannot be negative.
            If > 1, it will apply the transform floor(p) times and apply a last time with a probability of p - floor(p).
    """
    if p < 0.0:
        raise ValueError(f"Invalid argument {p=} < 0")
    if isinstance(generator, int):
        generator = Generator().manual_seed(generator)

    p_floor = math.floor(p)
    for _ in range(p_floor):
        x = transform(x)

    sampled = torch.rand((), generator=generator)
    if sampled + p_floor < p:
        x = transform(x)

    return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import (
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Mapping,
    Protocol,
    Sequence,
    Tuple,
    Union,
    runtime_checkable,
)

import torch
from torch import Tensor
from typing_extensions import TypeGuard

from torchoutil.utils.packaging import _NUMPY_AVAILABLE

if _NUMPY_AVAILABLE:
    import numpy as np


@runtime_checkable
class DataclassInstance(Protocol):
    # Class meant for typing purpose only
    __dataclass_fields__: ClassVar[Dict[str, Any]]


@runtime_checkable
class NamedTupleInstance(Protocol):
    # Class meant for typing purpose only
    _fields: Tuple[str, ...]
    _field_defaults: Dict[str, Any]

    def _asdict(self) -> Dict[str, Any]:
        ...

    def __getitem__(self, idx):
        ...

    def __len__(self) -> int:
        ...


BuiltinScalar = Union[int, float, bool, complex]


def is_numpy_scalar(x: Any) -> bool:
    """Returns True if x is an instance of a numpy generic type or a zero-dimensional numpy array.
    If numpy is not installed, this function always returns False.
    """
    if _NUMPY_AVAILABLE:
        return isinstance(x, np.generic) or (isinstance(x, np.ndarray) and x.ndim == 0)
    else:
        return False


def is_python_scalar(x: Any) -> TypeGuard[Union[BuiltinScalar]]:
    """Returns True if x is a builtin scalar type (int, float, bool, complex)."""
    return isinstance(x, (int, float, bool, complex))


def is_torch_scalar(x: Any) -> TypeGuard[Tensor]:
    """Returns True if x is a zero-dimensional torch Tensor."""
    return isinstance(x, Tensor) and x.ndim == 0


def is_scalar(x: Any) -> TypeGuard[Union[BuiltinScalar, Tensor]]:
    """Returns True if input is a scalar.

    Accepted scalars list is:
    - Python numbers (int, float, bool, complex)
    - PyTorch zero-dimensional tensors
    - Numpy zero-dimensional arrays
    - Numpy generic scalars
    """
    return (
        is_python_scalar(x)
        or is_torch_scalar(x)
        or (_NUMPY_AVAILABLE and is_numpy_scalar(x))
    )


def is_dataclass_instance(x: Any) -> TypeGuard[DataclassInstance]:
    return not isinstance(x, type) and isinstance(x, DataclassInstance)


def is_namedtuple_instance(x: Any) -> TypeGuard[NamedTupleInstance]:
    return not isinstance(x, type) and isinstance(x, NamedTupleInstance)


def is_dict_str(x: Any) -> TypeGuard[Dict[str, Any]]:
    return isinstance(x, dict) and all(isinstance(key, str) for key in x.keys())


def is_iterable_bool(
    x: Any,
    *,
    accept_tensor: bool = False,
) -> TypeGuard[Iterable[bool]]:
    return isinstance(x, Iterable) and (
        all(isinstance(xi, bool) for xi in x)
        or (
            accept_tensor
            and all(
                isinstance(xi, Tensor) and xi.ndim == 0 and xi.dtype == torch.bool
                for xi in x
            )
        )
    )


def is_sequence_bool(
    x: Any,
    *,
    accept_tensor: bool = False,
) -> TypeGuard[Sequence[bool]]:
    return isinstance(x, Sequence) and (
        all(isinstance(xi, bool) for xi in x)
        or (
            accept_tensor
            and all(
                isinstance(xi, Tensor) and xi.ndim == 0 and xi.dtype == torch.bool
                for xi in x
            )
        )
    )


def is_sequence_int(
    x: Any,
    *,
    accept_tensor: bool = False,
) -> TypeGuard[Sequence[int]]:
    return isinstance(x, Sequence) and (
        all(isinstance(xi, int) for xi in x)
        or (
            accept_tensor
            and all(
                isinstance(xi, Tensor)
                and xi.ndim == 0
                and not xi.is_floating_point()
                and not xi.is_complex()
                for xi in x
            )
        )
    )


def is_iterable_int(
    x: Any,
    *,
    accept_tensor: bool = False,
) -> TypeGuard[Iterable[int]]:
    return isinstance(x, Iterable) and (
        all(isinstance(xi, int) for xi in x)
        or (
            accept_tensor
            and all(
                isinstance(xi, Tensor)
                and xi.ndim == 0
                and not xi.is_floating_point()
                and not xi.is_complex()
                for xi in x
            )
        )
    )


def is_iterable_str(
    x: Any,
    *,
    accept_str: bool = True,
) -> TypeGuard[Iterable[str]]:
    return (accept_str and isinstance(x, str)) or (
        not isinstance(x, str)
        and isinstance(x, Iterable)
        and all(isinstance(xi, str) for xi in x)
    )


def is_sequence_str(
    x: Any,
    *,
    accept_str: bool = True,
) -> TypeGuard[Sequence[str]]:
    return (accept_str and isinstance(x, str)) or (
        not isinstance(x, str)
        and isinstance(x, Sequence)
        and all(isinstance(xi, str) for xi in x)
    )


def is_iterable_bytes_list(x: Any) -> TypeGuard[Iterable[Union[bytes, list]]]:
    return isinstance(x, Iterable) and all(isinstance(xi, (bytes, list)) for xi in x)


def is_iterable_iterable_int(x: Any) -> TypeGuard[Iterable[Iterable[int]]]:
    return (
        isinstance(x, Iterable)
        and all(isinstance(xi, Iterable) for xi in x)
        and all(isinstance(xij, int) for xi in x for xij in xi)
    )


def is_iterable_tensor(x: Any) -> TypeGuard[Iterable[Tensor]]:
    return isinstance(x, Iterable) and all(isinstance(xi, Tensor) for xi in x)


def is_list_list_str(x: Any) -> TypeGuard[List[List[str]]]:
    return (
        isinstance(x, list)
        and all(isinstance(xi, list) for xi in x)
        and all(isinstance(xij, str) for xi in x for xij in xi)
    )


def is_list_bool(x: Any) -> TypeGuard[List[bool]]:
    return isinstance(x, list) and all(isinstance(xi, bool) for xi in x)


def is_list_int(x: Any) -> TypeGuard[List[int]]:
    return isinstance(x, list) and all(isinstance(xi, int) for xi in x)


def is_list_str(x: Any) -> TypeGuard[List[str]]:
    return isinstance(x, list) and all(isinstance(xi, str) for xi in x)


def is_list_tensor(x: Any) -> TypeGuard[List[Tensor]]:
    return isinstance(x, list) and all(isinstance(xi, Tensor) for xi in x)


def is_mapping_str(x: Any) -> TypeGuard[Mapping[str, Any]]:
    return isinstance(x, Mapping) and all(isinstance(key, str) for key in x.keys())


def is_tuple_tensor(x: Any) -> TypeGuard[Tuple[Tensor, ...]]:
    return isinstance(x, tuple) and all(isinstance(xi, Tensor) for xi in x)


def is_tuple_str(x: Any) -> TypeGuard[Tuple[str, ...]]:
    return isinstance(x, tuple) and all(isinstance(xi, str) for xi in x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import io
from pathlib import Path
from typing import Any, Dict, List, Literal, Mapping, Sequence, Union, overload

from torchoutil.utils.collections import dict_list_to_list_dict, list_dict_to_dict_list
from torchoutil.utils.saving.common import to_builtin


def to_csv(
    data: Union[Sequence[Mapping[str, Any]], Mapping[str, Sequence[Any]]],
    fpath: Union[str, Path, None],
    *,
    overwrite: bool = True,
    to_builtins: bool = False,
    make_parents: bool = True,
    **csv_writer_kwargs,
) -> str:
    if isinstance(data, Mapping):
        data = dict_list_to_list_dict(data)
    else:
        data = [dict(data_i.items()) for data_i in data]

    if len(data) <= 0:
        raise ValueError(f"Invalid argument {data=}. (found empty iterable)")

    if fpath is not None:
        fpath = Path(fpath).resolve().expanduser()
        if not overwrite and fpath.exists():
            raise FileExistsError(f"File {fpath} already exists.")
        elif make_parents:
            fpath.parent.mkdir(parents=True, exist_ok=True)

    if to_builtins:
        data = to_builtin(data)

    fieldnames = list(data[0].keys())  # type: ignore
    file = io.StringIO()
    writer = csv.DictWriter(file, fieldnames, **csv_writer_kwargs)
    writer.writeheader()
    writer.writerows(data)  # type: ignore
    content = file.getvalue()
    file.close()

    if fpath is not None:
        fpath.write_text(content)
    return content


@overload
def load_csv(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["dict"],
    **csv_reader_kwargs,
) -> Dict[str, List[Any]]:
    ...


@overload
def load_csv(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["list", "dict"] = "list",
    **csv_reader_kwargs,
) -> List[Dict[str, Any]]:
    ...


def load_csv(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["list", "dict"] = "list",
    **csv_reader_kwargs,
) -> Union[List[Dict[str, Any]], Dict[str, List[Any]]]:
    with open(fpath, "r") as file:
        reader = csv.DictReader(file, **csv_reader_kwargs)
        data = list(reader)

    if orient == "dict":
        data = list_dict_to_dict_list(data, key_mode="same")
    elif orient == "list":
        pass
    else:
        ORIENT_VALUES = ("list", "dict")
        raise ValueError(
            f"Invalid argument {orient=}. (expected one of {ORIENT_VALUES})"
        )

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import logging
import re
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    List,
    Literal,
    Optional,
    OrderedDict,
    Protocol,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import torch
from torch import Tensor, nn
from torch.nn.parameter import Parameter

from torchoutil.nn.functional.others import count_parameters
from torchoutil.utils.type_checks import is_dict_str, is_mapping_str

InType = TypeVar("InType", covariant=False, contravariant=True)
OutType = TypeVar("OutType", covariant=True, contravariant=False)
OutType2 = TypeVar("OutType2", covariant=True, contravariant=False)

DEVICE_DETECT_MODES = ("proxy", "first_param", "none")
DeviceDetectMode = Literal["proxy", "first_param", "none"]


pylog = logging.getLogger(__name__)


class SupportsTypedForward(Protocol[InType, OutType]):
    def forward(self, x: InType) -> OutType:
        ...


class ProxyDeviceModule(nn.Module):
    def __init__(
        self,
        *,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        if device_detect_mode not in DEVICE_DETECT_MODES:
            raise ValueError(
                f"Invalid argument {device_detect_mode=}. (expected one of {DEVICE_DETECT_MODES})"
            )

        super().__init__()
        self.__device_detect_mode = device_detect_mode
        self.register_buffer("__proxy", torch.empty((0,)), persistent=False)

    @property
    def device_detect_mode(self) -> str:
        return self.__device_detect_mode

    def get_device(self) -> Optional[torch.device]:
        """Returns the Module device according to device_detect_mode property."""
        if self.__device_detect_mode == "proxy":
            return self._buffers["__proxy"].device  # type: ignore
        elif self.__device_detect_mode == "first_param":
            try:
                device0 = next(iter(self.get_devices(params=True, buffers=False)))
                return device0
            except StopIteration:
                return None
        else:
            return None

    def get_devices(
        self,
        *,
        params: bool = True,
        buffers: bool = True,
        recurse: bool = True,
    ) -> Iterator[torch.device]:
        """Returns an iterator over all unique devices in module."""
        its: List[Iterator[Union[Tensor, Parameter]]] = []
        if params:
            its.append(self.parameters(recurse=recurse))
        if buffers:
            its.append(self.buffers(recurse=recurse))

        devices = {}
        for it in its:
            for param_or_buffer in it:
                device = param_or_buffer.device
                if device not in devices:
                    yield device
                devices[param_or_buffer.device] = None


class ConfigModule(nn.Module):
    _CONFIG_EXCLUDE: ClassVar[Tuple[str, ...]] = tuple(
        f".*{k}" for k in nn.Module().__dict__.keys()
    ) + ("_.*",)
    _CONFIG_TYPES: ClassVar[Tuple[type, ...]] = (int, str, bool, float)

    def __init__(self, *, strict_load: bool = False) -> None:
        object.__setattr__(self, f"_{ConfigModule.__name__}__config", {})
        object.__setattr__(self, f"_{ConfigModule.__name__}__strict_load", strict_load)
        super().__init__()
        self.__config: Dict[str, Any]
        self.__strict_load: bool

    @property
    def config(self) -> Dict[str, Any]:
        return self.__config

    def __setattr__(self, name: str, value: Any) -> None:
        self.__update_config(name, value)
        return super().__setattr__(name, value)

    def __delattr__(self, name) -> None:
        self.__config.pop(name, None)
        return super().__delattr__(name)

    def add_module(self, name: str, module: Union[nn.Module, None]) -> None:
        self.__update_config(name, module)
        return super().add_module(name, module)

    def get_extra_state(self) -> Any:
        state = {"config": self.__config}
        return state

    def set_extra_state(self, state: Any) -> None:
        in_config = state["config"]
        if self.config == in_config:
            return None

        msg = f"Invalid saved config {in_config} with current one {self.config}."
        if self.__strict_load:
            raise ValueError(msg)
        else:
            pylog.warning(msg)

    def __update_config(self, name: str, value: Any) -> None:
        prefix = f"{name}."
        if self._is_config_value(name, value):
            subconfig = {name: value}
            prefix = ""
        elif isinstance(value, ConfigModule):
            subconfig = value.config
        elif hasattr(value, "_hparams") and is_mapping_str(value):
            subconfig = dict(value._hparams.items())  # type: ignore
        elif hasattr(value, "__dict__"):
            subconfig = value.__dict__
        else:
            subconfig = {}

        subconfig = {f"{prefix}{k}": v for k, v in subconfig.items()}
        subconfig = {k: v for k, v in subconfig.items() if self._is_config_value(k, v)}
        self.__config.update(subconfig)

    @classmethod
    def _is_config_value(cls, name: str, value: Any) -> bool:
        return isinstance(value, cls._CONFIG_TYPES) and all(
            re.match(exclude_i, name) is None for exclude_i in cls._CONFIG_EXCLUDE
        )


class TypedModule(Generic[InType, OutType], nn.Module):
    """Typed version of torch.nn.Module. Can specify an input and output type."""

    def __call__(self, *args: InType, **kwargs: InType) -> OutType:
        return super().__call__(*args, **kwargs)

    def forward(self, *args: InType, **kwargs: InType) -> OutType:
        return super().forward(*args, **kwargs)

    @overload
    def compose(
        self,
        other: "TypedModule[Any, OutType2]",
    ) -> "TypedSequential[InType, OutType2]":
        ...

    @overload
    def compose(self, other: nn.Module) -> "TypedSequential[InType, Any]":
        ...

    def compose(self, other) -> "TypedSequential[InType, Any]":
        return TypedSequential(self, other)


class TypedSequential(
    Generic[InType, OutType],
    TypedModule[InType, OutType],
    nn.Sequential,
):
    def __init__(
        self,
        *args,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        TypedModule.__init__(self)
        nn.Sequential.__init__(self, *args)

        self.__unpack_tuple = unpack_tuple
        self.__unpack_dict = unpack_dict

    @property
    def unpack_tuple(self) -> bool:
        return self.__unpack_tuple

    @property
    def unpack_dict(self) -> bool:
        return self.__unpack_dict

    def __call__(self, x: InType) -> OutType:
        return nn.Sequential.__call__(self, x)

    def forward(self, x: InType) -> OutType:
        for module in self:
            if self.__unpack_tuple and isinstance(x, tuple):
                x = module(*x)
            elif self.__unpack_dict and is_dict_str(x):
                x = module(**x)
            else:
                x = module(x)
        return x  # type: ignore

    def tolist(self) -> List[nn.Module]:
        return list(self._modules.values())

    def todict(self) -> Dict[str, nn.Module]:
        return copy.copy(self._modules)


class EModule(
    Generic[InType, OutType],
    ConfigModule,
    TypedModule[InType, OutType],
    ProxyDeviceModule,
):
    """Enriched torch.nn.Module with proxy device, forward typing and automatic configuration detection from attributes.

    The default behaviour is the same than PyTorch Module class.
    """

    def __init__(
        self,
        *,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        # ConfigModule must be first
        ConfigModule.__init__(self, strict_load=strict_load)
        TypedModule.__init__(self)
        ProxyDeviceModule.__init__(self, device_detect_mode=device_detect_mode)

    def count_parameters(
        self,
        *,
        recurse: bool = True,
        only_trainable: bool = False,
        buffers: bool = False,
    ) -> int:
        """Returns the number of parameters in this module."""
        return count_parameters(
            self,
            recurse=recurse,
            only_trainable=only_trainable,
            buffers=buffers,
        )


TypedModuleLike = Union[
    SupportsTypedForward[InType, OutType],
    TypedModule[InType, OutType],
]


class ESequential(
    Generic[InType, OutType],
    EModule[InType, OutType],
    TypedSequential[InType, OutType],
):
    """Enriched torch.nn.Sequential with proxy device, forward typing and automatic configuration detection from attributes.

    Designed to work with `torchoutil.nn.EModule` instances.
    The default behaviour is the same than PyTorch Sequential class.
    """

    @overload
    def __init__(
        self,
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, Any],
        arg3: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, Any],
        arg3: TypedModuleLike[Any, Any],
        arg4: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, Any],
        arg3: TypedModuleLike[Any, Any],
        arg4: TypedModuleLike[Any, Any],
        arg5: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, Any],
        arg3: TypedModuleLike[Any, Any],
        arg4: TypedModuleLike[Any, Any],
        arg5: TypedModuleLike[Any, Any],
        arg6: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg: "OrderedDict[str, TypedModuleLike[InType, OutType]]",
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg: "OrderedDict[str, nn.Module]",
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        *args: nn.Module,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        ...

    def __init__(
        self,
        *args,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        device_detect_mode: DeviceDetectMode = "proxy",
    ) -> None:
        EModule.__init__(
            self,
            strict_load=strict_load,
            device_detect_mode=device_detect_mode,
        )
        TypedSequential.__init__(
            self,
            *args,
            unpack_tuple=unpack_tuple,
            unpack_dict=unpack_dict,
        )


def __test_typing_1() -> None:
    import torch
    from torch import Tensor

    class LayerA(TypedModule[Tensor, Tensor]):
        def forward(self, x: Tensor) -> Tensor:
            return x * x

    class LayerB(TypedModule[Tensor, int]):
        def forward(self, x: Tensor) -> int:
            return int(x.sum().item())

    class LayerC(TypedModule[int, Tensor]):
        def forward(self, x: int) -> Tensor:
            return torch.as_tensor(x)

    x = torch.rand(10)
    xa = LayerA()(x)
    xb = LayerB()(x)

    seq = ESequential(LayerA(), LayerA(), LayerB())
    xab = seq(x)

    seq = LayerA().compose(LayerA()).compose(LayerB())
    xab = seq(x)

    seq = LayerC().compose(LayerA())
    xc = seq(2)

    assert isinstance(xa, Tensor)
    assert isinstance(xb, int)
    assert isinstance(xab, int)
    assert isinstance(xc, Tensor)

    class LayerD(nn.Module):
        def forward(self, x: Tensor) -> int:
            return int(x.item())

    class LayerE(nn.Module):
        def forward(self, x: bool) -> str:
            return str(x)

    seq = ESequential(LayerD(), LayerE())
    y = seq(torch.rand())

    assert isinstance(y, str)

    class LayerF(TypedModule[bool, str]):
        def forward(self, x):
            return str(x)

    seq = ESequential(LayerF())
    y = seq(True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import sys
from pathlib import Path
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar, Union

import torch
from torch.utils.data.dataset import Dataset

from torchoutil.utils.pack.common import (
    ATTRS_FNAME,
    ContentMode,
    PackedDatasetAttributes,
)

pylog = logging.getLogger(__name__)

T = TypeVar("T")
U = TypeVar("U")


class PackedDataset(Generic[T, U], Dataset[U]):
    def __init__(
        self,
        root: Union[str, Path],
        transform: Optional[Callable[[T], U]] = None,
        *,
        load_fn: Callable[..., Union[T, List[T]]] = torch.load,
        load_kwds: Optional[Dict[str, Any]] = None,
        use_cache: bool = False,
    ) -> None:
        """

        Args:
            root: Root directory containing the info.json and the data files.
            transform: Optional transform to apply to each item. defaults to None.
            load_fn: Load function to load an item or batch. defaults to torch.load.
            load_kwds: Keywords arguments passed to load_fn. defaults to None.
            use_cache: If True, cache each item or batch in memory. defaults to False.
        """
        root = Path(root)
        if load_kwds is None:
            load_kwds = {}

        super().__init__()
        self._root = root
        self._transform = transform
        self._load_fn = load_fn
        self._load_kwds = load_kwds
        self._use_cache = use_cache

        self._attrs = {}
        self._fpaths = []
        self._cache = None
        self._cache_idx = None
        self._reload_data()

    @property
    def attrs(self) -> PackedDatasetAttributes:
        return self._attrs  # type: ignore

    @property
    def content_mode(self) -> Optional[ContentMode]:
        return self._attrs.get("content_mode")

    @property
    def batch_size(self) -> int:
        return self._attrs["batch_size"]

    def __getitem__(self, idx: int) -> U:
        item = self._load_item(idx)
        if self._transform is not None:
            item = self._transform(item)
        return item  # type: ignore

    def __len__(self) -> int:
        return self._attrs["length"]

    def _load_item(self, idx: int) -> T:
        if self.content_mode == "item":
            batch_size = 1
        elif self.content_mode == "batch":
            batch_size = self.batch_size
        else:
            raise RuntimeError(
                f"Invalid PickleDataset state. (cannot load item with {self.content_mode=})"
            )

        target_cache_idx = idx // batch_size
        if self._cache is None or self._cache_idx != target_cache_idx:
            self._cache = None
            path = self._fpaths[target_cache_idx]
            item_or_batch = self._load_fn(path, **self._load_kwds)

            if self._use_cache:
                self._cache = item_or_batch
                self._cache_idx = target_cache_idx
        else:
            item_or_batch = self._cache

        if self.content_mode == "item":
            item: T = item_or_batch  # type: ignore
        else:
            batch: List[T] = item_or_batch  # type: ignore
            local_idx = idx % batch_size
            item = batch[local_idx]

        return item

    def _reload_data(self) -> None:
        attrs_fpath = self._root.joinpath(ATTRS_FNAME)

        if not attrs_fpath.is_file():
            raise FileNotFoundError(f"Cannot find attribute file '{str(attrs_fpath)}'.")

        with open(attrs_fpath, "r") as file:
            attrs = json.load(file)

        # Disable check for python <= 3.8 because __required_keys__ does not exists in this version
        if sys.version_info.major < 3 or (
            sys.version_info.major == 3 and sys.version_info.minor <= 8
        ):
            missing = []
        else:
            missing = list(
                set(PackedDatasetAttributes.__required_keys__).difference(attrs)
            )

        if len(missing) > 0:
            raise RuntimeError(
                f"Missing {len(missing)} keys in attribute file. (with {missing=})"
            )

        content_dname = attrs["content_dname"]
        content_dpath = self._root.joinpath(content_dname)

        fnames = attrs["files"]
        fpaths = [content_dpath.joinpath(fname) for fname in fnames]

        self._attrs = attrs
        self._fpaths = fpaths

    @classmethod
    def is_pickle_root(cls, root: Union[str, Path]) -> bool:
        try:
            PackedDataset(root)
            return True
        except (FileNotFoundError, RuntimeError):
            return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.functional.transform import (
    repeat_interleave_nd,
    resample_nearest_rates,
)


class TestRepeat(TestCase):
    def test_example_1(self) -> None:
        x = torch.as_tensor([[0, 1, 2, 3], [4, 5, 6, 7]])
        result = repeat_interleave_nd(x, repeats=2, dim=0)
        expected = torch.as_tensor(
            [[0, 1, 2, 3], [0, 1, 2, 3], [4, 5, 6, 7], [4, 5, 6, 7]]
        )
        assert torch.equal(result, expected)


class TestResampleNearest(TestCase):
    def test_example_1(self) -> None:
        x = torch.arange(10, 20)
        result = resample_nearest_rates(x, 0.5)
        expected = torch.as_tensor([10, 12, 14, 16, 18])
        assert torch.equal(result, expected)

    def test_example_2(self) -> None:
        x = torch.arange(10, 20)
        result = resample_nearest_rates(x, 2)
        expected = torch.as_tensor(
            [
                10,
                10,
                11,
                11,
                12,
                12,
                13,
                13,
                14,
                14,
                15,
                15,
                16,
                16,
                17,
                17,
                18,
                18,
                19,
                19,
            ]
        )
        assert torch.equal(result, expected)

    def test_example_3(self) -> None:
        x = torch.stack([torch.arange(10, 20), torch.arange(20, 30)])
        result = resample_nearest_rates(x, 0.5)
        expected = torch.as_tensor([[10, 12, 14, 16, 18], [20, 22, 24, 26, 28]])
        assert torch.equal(result, expected)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import sys
from functools import cache
from logging import FileHandler, Formatter, Logger, StreamHandler
from pathlib import Path
from types import ModuleType
from typing import IO, List, Literal, Optional, Sequence, Union

from torchoutil.utils.packaging import _COLORLOG_AVAILABLE

pylog = logging.getLogger(__name__)

DEFAULT_FMT = "[%(asctime)s][%(name)s][%(levelname)s] - %(message)s"
PackageOrLogger = Union[str, ModuleType, None, Logger]


if _COLORLOG_AVAILABLE:
    from colorlog import ColoredFormatter


@cache
def warn_once(msg: str, logger: PackageOrLogger) -> None:
    pylog = _get_loggers(logger)[0]
    pylog.warning(msg)


def setup_logging_verbose(
    package_or_logger: Union[
        PackageOrLogger,
        Sequence[PackageOrLogger],
    ],
    verbose: Optional[int],
    fmt: Union[str, None, Formatter] = DEFAULT_FMT,
    stream: Union[IO[str], Literal["auto"]] = "auto",
) -> None:
    if verbose is None:
        level = None
    else:
        level = _verbose_to_logging_level(verbose)
    return setup_logging_level(package_or_logger, level=level, fmt=fmt, stream=stream)


def setup_logging_level(
    package_or_logger: Union[
        PackageOrLogger,
        Sequence[PackageOrLogger],
    ],
    level: Optional[int],
    fmt: Union[str, None, Formatter] = DEFAULT_FMT,
    stream: Union[IO[str], Literal["auto"]] = "auto",
) -> None:
    logger_lst = _get_loggers(package_or_logger)
    if isinstance(fmt, str):
        fmt = Formatter(fmt)
    if stream == "auto":
        if running_on_interpreter():
            stream = sys.stdout
        else:
            stream = sys.stderr

    for logger in logger_lst:
        found = False

        for handler in logger.handlers:
            if isinstance(handler, StreamHandler) and handler.stream is stream:
                handler.setFormatter(fmt)
                found = True
                break

        if not found:
            handler = StreamHandler(stream)
            handler.setFormatter(fmt)
            logger.addHandler(handler)

        if level is not None:
            logger.setLevel(level)


def _get_loggers(
    package_or_logger: Union[
        PackageOrLogger,
        Sequence[PackageOrLogger],
    ],
) -> List[Logger]:
    if package_or_logger is None or isinstance(
        package_or_logger, (str, Logger, ModuleType)
    ):
        package_or_logger_lst = [package_or_logger]
    else:
        package_or_logger_lst = list(package_or_logger)

    name_or_logger_lst = [
        pkg.__name__ if isinstance(pkg, ModuleType) else pkg
        for pkg in package_or_logger_lst
    ]
    logger_lst = [
        logging.getLogger(pkg_i) if not isinstance(pkg_i, Logger) else pkg_i
        for pkg_i in name_or_logger_lst
    ]
    return logger_lst


def _verbose_to_logging_level(verbose: int) -> int:
    if verbose < 0:
        level = logging.ERROR
    elif verbose == 0:
        level = logging.WARNING
    elif verbose == 1:
        level = logging.INFO
    else:
        level = logging.DEBUG
    return level


class CustomFileHandler(FileHandler):
    """FileHandler that build intermediate directories.

    Used for export hydra logs to a file contained in a folder that does not exists yet at the start of the program.
    """

    def __init__(
        self,
        filename: Union[str, Path],
        mode: str = "a",
        encoding: Optional[str] = None,
        delay: bool = True,
        errors: Optional[str] = None,
    ) -> None:
        filename = Path(filename)
        filename.parent.mkdir(parents=True, exist_ok=True)

        super().__init__(filename, mode, encoding, delay, errors)


def running_on_interpreter() -> bool:
    return get_ipython_name() is None


def running_on_terminal() -> bool:
    return get_ipython_name() == "TerminalInteractiveShell"


def running_on_notebook() -> bool:
    return get_ipython_name() == "ZMQInteractiveShell"


def get_ipython_name() -> (
    Optional[Literal["TerminalInteractiveShell", "ZMQInteractiveShell"]]
):
    try:
        return get_ipython().__class__.__name__
    except NameError:
        return None


def get_colored_formatter() -> Formatter:
    if not _COLORLOG_AVAILABLE:
        raise RuntimeError(
            "Cannot call function get_colored_fmt() without colorlog installed. Please use `pip install torchoutil[extras]` to install it."
        )

    rank = os.getenv("SLURM_PROCID", 0)
    fmt = f"[%(purple)sRANK{rank}%(reset)s][%(cyan)s%(asctime)s%(reset)s][%(blue)s%(name)s%(reset)s][%(log_color)s%(levelname)s%(reset)s] - %(message)s"
    formatter = ColoredFormatter(
        fmt=fmt,
        log_colors={
            "DEBUG": "purple",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "bold_red",
        },
    )
    return formatter



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.modules import CropDim, PadDim, ResampleNearestRates
from torchoutil.nn.modules.mixins import ESequential


class TestTransforms(TestCase):
    def test_example_1(self) -> None:
        target_length = 10
        transform = ESequential(
            ResampleNearestRates(0.5),
            PadDim(target_length),
            ResampleNearestRates(2.5),
            CropDim(target_length),
        )

        x = torch.rand(10, 20, target_length)
        result = transform(x)

        assert x.shape == result.shape


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from argparse import Namespace
from collections import Counter, OrderedDict
from dataclasses import asdict
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Iterable, List, Literal, Mapping, TypeVar, Union, overload

from torch import Tensor
from torch.types import Number as TorchNumber

from torchoutil.utils.packaging import (
    _NUMPY_AVAILABLE,
    _OMEGACONF_AVAILABLE,
    _PANDAS_AVAILABLE,
)
from torchoutil.utils.type_checks import is_dataclass_instance, is_namedtuple_instance

if _NUMPY_AVAILABLE:
    import numpy as np

if _OMEGACONF_AVAILABLE:
    from omegaconf import DictConfig, ListConfig, OmegaConf

if _PANDAS_AVAILABLE:
    from pandas import DataFrame


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

UnkMode = Literal["identity", "error"]
UNK_MODES = ("identity", "error")


@overload
def to_builtin(
    x: Enum,
    *,
    unk_mode: UnkMode = "identity",
) -> str:
    ...


@overload
def to_builtin(
    x: Path,
    *,
    unk_mode: UnkMode = "identity",
) -> str:
    ...


@overload
def to_builtin(
    x: Namespace,
    *,
    unk_mode: UnkMode = "identity",
) -> Dict[str, Any]:
    ...


@overload
def to_builtin(
    x: Tensor,
    *,
    unk_mode: UnkMode = "identity",
) -> Union[List, TorchNumber]:
    ...


@overload
def to_builtin(
    x: Mapping[K, V],
    *,
    unk_mode: UnkMode = "identity",
) -> Dict[K, V]:
    ...


@overload
def to_builtin(
    x: T,
    *,
    unk_mode: UnkMode = "identity",
) -> T:
    ...


def to_builtin(
    x: Any,
    *,
    unk_mode: UnkMode = "identity",
) -> Any:
    """Helper function to sanitize data before saving to YAML or CSV file.

    Args:
        x: Object to convert to built-in equivalent.
        unk_mode: When an object type is not recognized, unk_mode defines the behaviour.
            If unk_mode == "identity", the object is returned unchanged.
            If unk_mode == "error", a TypeError is raised.
    """
    # Terminal cases
    if isinstance(x, (int, float, bool, complex, str, bytes)):
        return x
    elif isinstance(x, Enum):
        return x.name
    elif isinstance(x, Path):
        return str(x)
    elif isinstance(x, Tensor):
        return x.tolist()
    elif _NUMPY_AVAILABLE and isinstance(x, np.ndarray):
        return x.tolist()
    elif _NUMPY_AVAILABLE and isinstance(x, np.generic):
        return x.item()
    # Non-terminal cases
    elif _OMEGACONF_AVAILABLE and isinstance(x, (DictConfig, ListConfig)):
        return to_builtin(OmegaConf.to_container(x, resolve=False, enum_to_str=True))
    elif _PANDAS_AVAILABLE and isinstance(x, DataFrame):
        return to_builtin(x.to_dict())
    elif isinstance(x, Namespace):
        return to_builtin(x.__dict__)
    elif isinstance(x, (Counter, OrderedDict)):
        return to_builtin(dict(x))
    elif is_dataclass_instance(x):
        return to_builtin(asdict(x))
    elif is_namedtuple_instance(x):
        return to_builtin(x._asdict())
    elif isinstance(x, Mapping):
        return {to_builtin(k): to_builtin(v) for k, v in x.items()}  # type: ignore
    elif isinstance(x, Iterable):
        return [to_builtin(xi) for xi in x]  # type: ignore
    elif unk_mode == "identity":
        return x
    elif unk_mode == "error":
        raise TypeError(f"Unsupported argument type {type(x)}.")
    else:
        raise ValueError(f"Invalid argument {unk_mode=}. (expected one of {UNK_MODES})")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchoutil.utils.packaging import _YAML_AVAILABLE

from .common import to_builtin
from .csv_io import load_csv, to_csv

if _YAML_AVAILABLE:
    from .yaml_io import load_yaml, to_yaml



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Sized,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import torch
from torch import Size, Tensor, nn
from typing_extensions import TypeGuard

from torchoutil.nn.functional.get import get_device
from torchoutil.nn.functional.numpy import ACCEPTED_NUMPY_DTYPES
from torchoutil.utils.collections import all_eq
from torchoutil.utils.packaging import _NUMPY_AVAILABLE
from torchoutil.utils.type_checks import (
    is_list_tensor,
    is_numpy_scalar,
    is_python_scalar,
    is_scalar,
    is_torch_scalar,
    is_tuple_tensor,
)

if _NUMPY_AVAILABLE:
    import numpy as np


T = TypeVar("T")
U = TypeVar("U")


def count_parameters(
    model: nn.Module,
    *,
    recurse: bool = True,
    only_trainable: bool = False,
    buffers: bool = False,
) -> int:
    """Returns the number of parameters in a module."""
    params = (
        param
        for param in model.parameters(recurse)
        if not only_trainable or param.requires_grad
    )

    if buffers:
        params = itertools.chain(params, (buffer for buffer in model.buffers(recurse)))

    num_params = sum(param.numel() for param in params)
    return num_params


def find(
    x: Tensor,
    value: Any,
    default: Union[None, Tensor, int, float] = None,
    dim: int = -1,
) -> Tensor:
    """Return the index of the first occurrence of value in a tensor."""
    if x.ndim == 0:
        raise ValueError(
            f"Function find does not supports 0-d tensors. (found {x.ndim=} == 0)"
        )
    mask = x.eq(value)
    contains = mask.any(dim=dim)
    indices = mask.int().argmax(dim=dim)

    if default is None:
        if not contains.all():
            raise RuntimeError(f"Cannot find {value=} in tensor.")
        return indices
    else:
        output = torch.where(contains, indices, default)
        return output


@overload
def move_to_rec(
    x: Mapping[T, U],
    predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    **kwargs,
) -> Dict[T, U]:
    ...


@overload
def move_to_rec(
    x: T,
    predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    **kwargs,
) -> T:
    ...


def move_to_rec(
    x: Any,
    predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    **kwargs,
) -> Any:
    """Move all modules and tensors recursively to a specific dtype or device."""
    if "device" in kwargs:
        kwargs["device"] = get_device(kwargs["device"])

    if isinstance(x, (Tensor, nn.Module)):
        if predicate is None or predicate(x):
            return x.to(**kwargs)
        else:
            return x
    elif isinstance(x, (str, float, int, bool, complex)):
        return x
    elif isinstance(x, Mapping):
        return {k: move_to_rec(v, predicate=predicate, **kwargs) for k, v in x.items()}
    elif isinstance(x, Iterable):
        generator = (move_to_rec(xi, predicate=predicate, **kwargs) for xi in x)
        if isinstance(x, Generator):
            return generator
        elif isinstance(x, tuple):
            return tuple(generator)
        else:
            return list(generator)
    else:
        return x


def can_be_stacked(
    tensors: Union[List[Any], Tuple[Any, ...]],
) -> TypeGuard[Union[List[Tensor], Tuple[Tensor, ...]]]:
    """Returns True if inputs can be passed to `torch.stack` function."""
    if not is_list_tensor(tensors) and not is_tuple_tensor(tensors):
        return False
    if len(tensors) == 0:
        return False
    shape0 = tensors[0].shape
    result = all(tensor.shape == shape0 for tensor in tensors[1:])
    return result


def can_be_converted_to_tensor(x: Any) -> bool:
    """Returns True if inputs can be passed to `torch.as_tensor` function.

    This function returns False for heterogeneous inputs like `[[], 1]`, but this kind of value can be accepted by `torch.as_tensor`.
    """
    if isinstance(x, Tensor):
        return True
    else:
        return __can_be_converted_to_tensor_nested(x)


def ndim(x: Any) -> int:
    valid, ndim = _search_ndim(x)
    if valid:
        return ndim
    else:
        raise ValueError(f"Invalid argument {x}. (cannot compute ndim)")


def shape(x: Any) -> Size:
    valid, shape = _search_shape(x)
    if valid:
        shape = Size(shape)
        return shape
    else:
        raise ValueError(f"Invalid argument {x}. (cannot compute shape)")


def item(x: Any) -> Union[int, float, bool, complex]:
    if is_python_scalar(x):
        return x
    elif is_torch_scalar(x) or is_numpy_scalar(x):
        return x.item()
    else:
        raise ValueError(f"Invalid argument {x=}. (expected scalar number object)")


def _search_ndim(x: Any) -> Tuple[bool, int]:
    if is_scalar(x) or isinstance(x, str):
        return True, 0
    elif isinstance(x, Tensor) or (_NUMPY_AVAILABLE and isinstance(x, np.ndarray)):
        return True, x.ndim
    elif isinstance(x, Iterable):
        ndims = [_search_ndim(xi)[1] for xi in x]
        if len(ndims) == 0:
            return True, 1
        elif all_eq(ndims):
            return True, ndims[0] + 1
        else:
            return False, -1
    else:
        raise TypeError(f"Invalid argument type {type(x)}.")


def _search_shape(x: Any) -> Tuple[bool, Tuple[int, ...]]:
    if is_scalar(x) or isinstance(x, str):
        return True, ()
    elif isinstance(x, Tensor) or (_NUMPY_AVAILABLE and isinstance(x, np.ndarray)):
        return True, tuple(x.shape)
    elif isinstance(x, Iterable):
        shapes = [_search_shape(xi)[1] for xi in x]
        if len(shapes) == 0:
            return True, (0,)
        elif all_eq(shapes):
            return True, (len(shapes),) + shapes[0]
        else:
            return False, ()
    else:
        raise TypeError(f"Invalid argument type {type(x)}.")


def __can_be_converted_to_tensor_list_tuple(x: Union[List, Tuple]) -> bool:
    if len(x) == 0:
        return True

    valid_items = all(__can_be_converted_to_tensor_nested(xi) for xi in x)
    if not valid_items:
        return False

    if all(
        (not isinstance(xi, Sized) or (isinstance(xi, Tensor) and xi.ndim == 0))
        for xi in x
    ):
        return True
    elif all(isinstance(xi, Sized) for xi in x):
        len0 = len(x[0])
        return all(len(xi) == len0 for xi in x[1:])
    else:
        return False


def __can_be_converted_to_tensor_nested(x: Any) -> bool:
    if is_python_scalar(x):
        return True
    elif isinstance(x, Tensor) and x.ndim == 0:
        return True
    elif (
        _NUMPY_AVAILABLE
        and isinstance(x, (np.ndarray, np.generic))
        and x.dtype in ACCEPTED_NUMPY_DTYPES
    ):
        return True
    elif isinstance(x, (List, Tuple)):
        return __can_be_converted_to_tensor_list_tuple(x)
    else:
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from argparse import Namespace
from pathlib import Path
from typing import Any, Mapping, Union

from yaml import MappingNode, Node, SafeLoader, ScalarNode, SequenceNode

from torchoutil.utils.packaging import _OMEGACONF_AVAILABLE, _YAML_AVAILABLE
from torchoutil.utils.saving.common import to_builtin
from torchoutil.utils.type_checks import DataclassInstance, NamedTupleInstance

if _YAML_AVAILABLE:
    import yaml
else:
    raise ImportError(
        "Cannot use to_yaml python module since pyyaml package is not installed."
    )

if _OMEGACONF_AVAILABLE:
    from omegaconf import OmegaConf


def to_yaml(
    data: Union[Mapping[str, Any], Namespace, DataclassInstance, NamedTupleInstance],
    fpath: Union[str, Path, None],
    *,
    overwrite: bool = True,
    to_builtins: bool = False,
    make_parents: bool = True,
    resolve: bool = False,
    sort_keys: bool = False,
    indent: Union[int, None] = None,
    **yaml_dump_kwargs,
) -> str:
    if not _OMEGACONF_AVAILABLE and resolve:
        raise ValueError(
            "Cannot resolve yaml config without omegaconf package."
            "Please use resolve=False or install omegaconf with 'pip install torchoutil[extras]'."
        )

    if fpath is not None:
        fpath = Path(fpath).resolve().expanduser()
        if not overwrite and fpath.exists():
            raise FileExistsError(f"File {fpath} already exists.")
        elif make_parents:
            fpath.parent.mkdir(parents=True, exist_ok=True)

    if resolve:
        data = OmegaConf.create(data)  # type: ignore
        data = OmegaConf.to_container(data, resolve=True)  # type: ignore

    if to_builtins:
        data = to_builtin(data)

    content = yaml.dump(data, sort_keys=sort_keys, indent=indent, **yaml_dump_kwargs)
    if fpath is not None:
        fpath.write_text(content, encoding="utf-8")
    return content


def load_yaml(fpath: Union[str, Path]) -> Any:
    with open(fpath, "r") as file:
        data = yaml.safe_load(file)
    return data


class IgnoreTagLoader(SafeLoader):
    """SafeLoader that ignores yaml tags.

    Usage:

    ```python
    >>> content = yaml.load(stream, Loader=IgnoreTagLoader)
    ```
    """

    def construct_with_tag(self, tag: str, node: Node) -> Any:
        if isinstance(node, MappingNode):
            return self.construct_mapping(node)
        elif isinstance(node, ScalarNode):
            return self.construct_scalar(node)
        elif isinstance(node, SequenceNode):
            return self.construct_sequence(node)
        else:
            raise NotImplementedError(f"Unsupported node type {type(node)}.")


IgnoreTagLoader.add_multi_constructor("!", IgnoreTagLoader.construct_with_tag)
IgnoreTagLoader.add_multi_constructor("tag:", IgnoreTagLoader.construct_with_tag)
